# RegistrationApp for Django: Register users, with e-mail verification and CAPTCHA.
#
# Copyright (C) 2008  Juergen Brendel
#
# Dual licensed under AGPL and BSD license.

# This file handles the generation of CAPTCHA texts and images.
# Use make_captcha_string() to create the actual string.
# Use generate_image() to create the CAPTCHA image.
#
# As a part of the RegistrationApp, this file imports reg_settings
#
# ACKNOWLEDGEMENT: This code borrows heavily from the CAPTCHA package
# provided by Martin Winkler, especially the image generation function.
# It has all been quite heavily modified, and I stripped out a lot of
# stuff, because my setup and framework here is somewhat different.
# Nevertheless, I didn't know anything about the PIL, but using Martin's
# code as a template I was able to take it from there.

import tempfile
from os import listdir, path
from random import choice, randrange, randint
from PIL import Image, ImageFont, ImageDraw, ImageOps, ImageEnhance

from reg_settings import *

FGCOLS            = [ '#000000' ]   # Each letter/line has random color. Specify only one to have it fixed...
BGCOLS            = [ '#ffffff' ]   # ... same with backgrounds.
MIN_CHAR_YPOS     = 8               # Where the letters can be on the Y axis
MAX_CHAR_YPOS     = 15
MAX_CHAR_ROT      = 10              # Letters are randomly rotated by plus/minus this amount
MIN_CHAR_HEIGHT   = 38              # The minimum and maximum size. It will look even more...
MAX_CHAR_HEIGHT   = 40              # ... chaotic, because we use different fonts.
MIN_CHAR_KERNING  = 1               # Can be negative to jam the letters closer together...
MAX_CHAR_KERNING  = 4               # ... this is a GREAT way to make it tough on bots.
NUM_LINES         = 0               # How many random background lines we will have...
LINE_WEIGHT       = 1               # ... and how thick they are. Not very effective.
WHITE_NUM_LINES   = 0               # How many lines in background color chross through the characters...
WHITE_LINE_WEIGHT = 2               # ... this can be tough for bots, especially with thicker lines.
IMAGESIZE_X       = 200             # Dimensions of the generated image
IMAGESIZE_Y       = 60
RAND_X_BORDER     = 20              # Lines don't all start at the edge of the image. They can...
RAND_Y_BORDER     = 20              # ... start randomly from 0 to this many pixels in.
JAG_LINE          = False           # A jagged line through the image is a VERY good way...
JAG_NUM_KINKS     = 4               # ... to make it tough on bots. Set how many segments...
JAG_WEIGHT        = 3               # ... you want that line to have and how thick it should be.
FUZZY_ITERS       = 2               # Set to higher to make picture more fuzzy (0 = None). Don't overdo it.
CHAR_SET          = "abcdefghkmnpqrstuvwxyz"  # Try to avoid letters/numbers that can look too similar, such as 1 and I
#CHAR_SET          = "234567823456782345678abcdefghijkmnprstuvwxyzABDEFGHJKLMNPQRSTUVWXYZ"
CAPTCHA_LEN       = 5               # How long should the CAPTCHA be. That's only on average!
CAPTCHA_FONT_DIR  = captcha_font_dir   # Absolute path to where the fonts are located.


# Calculated values. We use them frequently, so
# we only want to calculate them once.
MAXX              = IMAGESIZE_X-1
MAXY              = IMAGESIZE_Y-1
SECTION_WIDTH     = (MAXX-2*RAND_X_BORDER)/JAG_NUM_KINKS


def __draw_lines(draw, colors, num, width):
    """
    Draw a number of lines across the image.

    The callers specifies the color, the number of
    lines and their width. Note that num lines will
    be drawn from left to right, and then the same
    number again (!) from top to bottom. So, the
    actual number of lines is exactly twice that
    which is specified here.

    """
    for i in xrange(0,num):
        linepoints = [randint(0,MAXX), randint(0,RAND_Y_BORDER), randint(0,MAXX), randint(MAXY-RAND_Y_BORDER,MAXY)]
        draw.line(linepoints, width=width, fill=choice(colors))

        linepoints = [randint(0, RAND_X_BORDER), randint(0,MAXY), randint(MAXX-RAND_X_BORDER,MAXX), randint(0,MAXY)]
        draw.line(linepoints, width=width, fill=choice(colors))


def __draw_jagged_line(draw, colors):
    """
    Draw a jagged line through the image.

    A jagged line is very useful to fight automatic
    segmentation. The caller can specify the color
    of the line.

    """
    ylow       = RAND_Y_BORDER
    yhigh      = MAXY-ylow
    linepoints = []

    x = RAND_X_BORDER
    for i in xrange(0, JAG_NUM_KINKS+1):
        linepoints.append(randint(x-RAND_X_BORDER,x+RAND_X_BORDER))
        linepoints.append(randint(ylow,yhigh))
        x += SECTION_WIDTH

    draw.line(linepoints, width=JAG_WEIGHT, fill=choice(colors))


def make_captcha_string():
    """
    Create a CAPTCHA string.

    The length of the captcha string is determined 
    by the CAPTCHA_LEN variable.

    Return the generated string.

    """
    char_val  = ""
    len       = CAPTCHA_LEN if CAPTCHA_LEN < 9 else 9
    len += randrange(-1,2)     # Vary the length by 1 down or 1 up (or not at all)
    for i in xrange(len):
        char_val += choice(CHAR_SET)
    return char_val


def generate_image(char_val):
    """
    Create the CAPTCHA as an image file.

    Return the filename as which it was created.

    """
    bgcolor = choice(BGCOLS)
    bgimage = Image.new('RGB', (IMAGESIZE_X, IMAGESIZE_Y), bgcolor)
    posnew  = 7 + RAND_X_BORDER
    fontdir = CAPTCHA_FONT_DIR
    fontnames = [ path.join(fontdir, x) for x in listdir(fontdir) ]


    __draw_lines(ImageDraw.Draw(bgimage), FGCOLS, NUM_LINES, LINE_WEIGHT)
    if JAG_LINE:
        __draw_jagged_line(ImageDraw.Draw(bgimage), FGCOLS)

    # generate characters
    solution = char_val

    # render characters
    for c in solution:
        fgimage = Image.new('RGB', (IMAGESIZE_X, IMAGESIZE_Y), choice(FGCOLS))
        font = ImageFont.truetype(choice(fontnames), randrange(MIN_CHAR_HEIGHT, MAX_CHAR_HEIGHT))
        charimage = Image.new('L', font.getsize(' %s ' % c), '#000000')
        draw = ImageDraw.Draw(charimage)
        draw.text((0,0), ' %s' % c, font=font, fill='#ffffff')
        charimage = charimage.rotate(randrange(-MAX_CHAR_ROT, MAX_CHAR_ROT), expand=1, resample=Image.BICUBIC)
        charimage = charimage.crop(charimage.getbbox())
        maskimage = Image.new('L', (IMAGESIZE_X, IMAGESIZE_Y))
        ypos = randrange(MIN_CHAR_YPOS, MAX_CHAR_YPOS)
        maskimage.paste(charimage, 
                (posnew, ypos, 
                    charimage.size[0]+posnew, 
                    charimage.size[1]+ypos)
                )
        bgimage = Image.composite(fgimage, bgimage, maskimage)
        posnew += charimage.size[0] + randrange(MIN_CHAR_KERNING, MAX_CHAR_KERNING)

    __draw_lines(ImageDraw.Draw(bgimage), [bgcolor], WHITE_NUM_LINES, WHITE_LINE_WEIGHT)


    # Make the image smoother/fuzzier. Iterate more often to
    # increase fuzziness.
    for i in xrange(FUZZY_ITERS):
        enhancer = ImageEnhance.Sharpness(bgimage)
        bgimage = enhancer.enhance(0)
    # Enable this next line to see for a moment how easy it is
    # to get all the nice contrast back. So, don't overdo it with
    # the fuzziness. It will make it difficult for humans, but
    # not for computers.
    #bgimage = ImageOps.solarize(bgimage, threshold=254)

    # save file
    fname = tempfile.mktemp()
    try:
        bgimage.convert('P', palette=Image.ADAPTIVE, colors=32).save(fname + '.gif')
        ext = '.gif'
    except:
        bgimage.save(fname + '.jpg')
        ext = '.jpg'
    fname += ext
    return fname


