# RegistrationApp for Django: Register users, with e-mail verification and CAPTCHA.
#
# Copyright (C) 2008  Juergen Brendel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# The forms for the RegistrationApp.

from django.contrib.auth.forms import UserCreationForm
from django.contrib.auth.models import User
from django.core import validators
from django import oldforms
from django.utils.translation import gettext as _

import datetime

from models import EmailConfirmation, UserProfile, Captchas

import captcha


# This is a special validator for CAPTCHA values.
class MatchesCaptchaKey(object):
    """
    Validate CAPTCHA strings.

    Look up the correct value in the DB and compare.

    """
    def __init__(self, captcha_key_field_name, error_message=None):
        self.captcha_key = captcha_key_field_name
        self.error_message = error_message or lazy_inter(gettext_lazy("Incorrect."), self.other)
        self.always_test = True

    def __call__(self, field_data, all_data):
        # Get the stored characters for the captcha from the database
        key = all_data[self.captcha_key]
        try:
            captcha = Captchas.objects.get(key=key)
        except:
            raise validators.ValidationError, self.error_message
        cap_val = captcha.value.lower()
        captcha.delete()  # Once a CAPTCHA has been looked up, we can and should delete it.
        if field_data.lower() != cap_val:
            raise validators.ValidationError, self.error_message


# This is not terribly great, but it should work: I have copied
# the code from the UserCreationForm here, and have extended it
# to also contain an e-mail address field.
class ExtendedUserCreationForm(oldforms.Manipulator):
    "A form that creates a user, with no privileges, from the given username and password."
    def __init__(self):
        self.fields = (
            oldforms.TextField(field_name='username', length=30, maxlength=30, is_required=True,
                validator_list=[validators.isAlphaNumeric, self.isValidUsername]),
            oldforms.EmailField(field_name='email1', length=30, maxlength=80, is_required=True),
            oldforms.EmailField(field_name='email2', length=30, maxlength=80, is_required=True,
                validator_list=[validators.AlwaysMatchesOtherField('email1', _("The two email fields didn't match."))]),
            oldforms.PasswordField(field_name='password1', length=30, maxlength=60, is_required=True),
            oldforms.PasswordField(field_name='password2', length=30, maxlength=60, is_required=True,
                validator_list=[validators.AlwaysMatchesOtherField('password1', _("The two password fields didn't match."))]),
            oldforms.HiddenField(field_name="key", is_required=True),
            oldforms.TextField(field_name='captcha', length=30, maxlength=30, is_required=True,
                validator_list=[MatchesCaptchaKey('key', _("The CAPTCHA value you entered is incorrect."))]),
        )

    def isValidUsername(self, field_data, all_data):
        """A validator to check if the username is acceptable."""
        try:
            User.objects.get(username=field_data)
        except User.DoesNotExist:
            return
        raise validators.ValidationError, _('A user with that username already exists.')

    def save(self, new_data):
        """
        Create the user.

        This also creates the matching profile entry for the user.
        
        """
        u = User.objects.create_user(username=new_data['username'], email=new_data['email1'], password=new_data['password1'])
        u.save()
        prof = UserProfile.objects.create(user=u, confirmation_time=datetime.datetime.now(), is_confirmed=0)
        prof.save()
        return u


# The form to change your e-mail address.
class EmailChangeForm(oldforms.Manipulator):
    def __init__(self):
        self.fields = (
            oldforms.EmailField(field_name='email1', length=30, maxlength=80, is_required=True),
            oldforms.EmailField(field_name='email2', length=30, maxlength=80, is_required=True,
                validator_list=[validators.AlwaysMatchesOtherField('email1', _("The two email fields didn't match."))]),
        )

