# RegistrationApp for Django: Register users, with e-mail verification and CAPTCHA.
#
# Copyright (C) 2008  Juergen Brendel
#
# Dual licensed under AGPL and BSD license.

# The views for the RegistrationApp.

import smtplib
import os

from django import oldforms
from django.http import HttpResponseRedirect, HttpResponse
from django.template import RequestContext
from django.shortcuts import render_to_response
from django.contrib import auth
from django.contrib.auth.models import User
from django.db.models import Q
from django.core.mail import send_mail

from models import EmailConfirmation, UserProfile, Captchas

from reg_settings import *
from reg_log import logger
from reg_helpers import *
from reg_forms import ExtendedUserCreationForm, EmailChangeForm

import captcha


# Django apparently didn't define this one? It looks like it only
# defined the Http404. Well, a 500 comes in handy, so we define it
# here.
class Http500(Exception):
    pass


# --------------------------------
# The view functions
# --------------------------------

@reg_login_required
def edit_settings(req):
    """
    Edit settings screen.

    Here a user can jump to screens to change things related to
    his/her registration profile. For example e-mail or password.

    """
    return render_to_response('edit_settings.html', None, context_instance=RequestContext(req))


@reg_login_required
@reg_confirmed_required
def email_change(req):
    """
    Change a user e-mail.

    This is the form that allows the user to change his/her registered
    e-mail address.

    """
    form = EmailChangeForm()
    if req.method == 'POST':
        data = req.POST.copy()
        errors = form.get_validation_errors(data)
        if not errors:
            new_email = data['email1']
            if new_email.lower() == req.user.email.lower():
                # We don't need to do anything if the e-mail wasn't actually changed.
                return HttpResponseRedirect(reg_prefix + "/email_change/done?status=equal")
            try:
                # We are sending confirmation e-mails to both the old and new address.
                msg = email_msg_address_changed % ( req.user.email, new_email )
                req.user.email_user(email_sub_address_changed, msg, site_email)
                send_mail(email_sub_address_changed, msg, site_email, [new_email])
            except:
                return HttpResponseRedirect(reg_prefix + "/email_change/done?status=failed")
                
            req.user.email = new_email
            req.user.save()
            return HttpResponseRedirect(reg_prefix + "/email_change/done")
    else:
        data, errors = {}, {}

    return render_to_response("email_change_form.html", { 'form' : oldforms.FormWrapper(form, data, errors) }, context_instance=RequestContext(req))


@reg_login_required
@reg_confirmed_required
def email_change_done(req):
    """
    Confirm the successful change of the e-mail.

    """
    status = req.GET.get('status', '')
    if status in ['failed', 'equal']:
        param = dict(status=status)
    else:
        param = {}
    return render_to_response("email_change_done.html", param, context_instance=RequestContext(req))


@reg_logout_required
def register(req):
    """
    Create a new user.

    A user creation form is displayed, including CAPTCHA.

    If the form is filled out correctly, then a confirmation
    e-mail is sent to the specified user address. Should an
    exception have happened when we try to send the e-mail,
    we are manually creating an error entry for the correct
    form field.

    The user account is not 'confirmed' until the user has
    clicked on the link that was sent in the confirmation
    e-mail. However, we want the users to be able to already
    browse the site a bit and get interested while they wait
    for the confirmation e-mail to arrive. That's why we log
    in the user immediately after completion of this form.
    But a flag is set to false in the user's profile (is_confirmed).
    This is made available to the forms and we also have a helper
    function for the testing of this in reg_helpers. So, the
    forms and the views can do special things when the user
    is not confirmed.

    """
    form = ExtendedUserCreationForm()
    new_user = None
    if req.method == 'POST':
        data = req.POST.copy()
        errors = form.get_validation_errors(data)
        if not errors:
            new_user = form.save(data)
            ec = None
            try:
                ec = EmailConfirmation.factory(new_user)
            except smtplib.SMTPException, e:
                errors['email1'] = ['Cannot send e-mail to this address.']
            except Exception, e:
                logger.error("Cannot create user account: " + str(e))
                errors['username'] = ['Cannot create new user account for some reason. Please try again later.']
            else:
                if allow_unconfirmed:
                    user = auth.authenticate(username=data['username'], password=data['password1'])
                    auth.login(req, user)
                return HttpResponseRedirect(reg_prefix + "/registration_email_sent")
    else:
        data, errors = {}, {}

    captcha = Captchas.factory()
    data['key'] = captcha.key

    if errors and new_user:
        new_user.delete()

    return render_to_response("register.html", {
        'form' : oldforms.FormWrapper(form, data, errors),
        'captcha_key' : captcha.key
    }, context_instance=RequestContext(req))


@reg_login_required
def registration_email_sent(req):
    """
    Display a notice about the confirmation e-mail having been sent correctly.

    """
    return render_to_response('registration_email_sent.html', param, context_instance=RequestContext(req))


def confirm_registration(req):
    """
    Check the supplied link against the email confirmation DB table.
    
    When the confirmation e-mails were sent out, a random ID was created.
    This is contained as a URL parameter in this request. We can take
    that key and check it against the DB to see if it exists, amd which
    user this should confirm.
    
    """
    key = req.GET.get('key', '')

    results = []
    if key:
        # See if we can find that key in the confirmation database
        qset = ( Q(key=key) & Q(is_used_up=0) )
        results = EmailConfirmation.objects.filter(qset)
        if len(results) > 1:
            logger.error("Multiple email confirmations for key: %s" % key)
            results= []

    if not results:
        return render_to_response('reg_conf_error.html', None, context_instance=RequestContext(req))
    else:
        ec = results[0]
        user_id = ec.user_id
        user = User.objects.get(id=user_id)
        if not user:
            logger.error("Confirmation request for non existing user. Key='%s', user-id='%s'" % (key, user_id))
            return render_to_response('reg_conf_error.html', None, context_instance=RequestContext(req))
        u_profile = user.get_profile()
        u_profile.is_confirmed = True
        u_profile.save()
        ec.used_up()
        return render_to_response('reg_conf_success.html', None, context_instance=RequestContext(req))


def get_captcha(req):
    """
    Returns the data for a captcha image.

    """
    key = req.GET.get('key', '')
    try:
        cap = Captchas.objects.get(key=key)
        cap_val = cap.value
        # We will temporarily create the CAPTCHA in an image file...
        fname = captcha.generate_image(cap_val)
        image_data = open(fname, "rb").read()
        os.remove(fname)   # ... which is now not needed anymore
        return HttpResponse(image_data, mimetype="image/png")
    except Exception, e:
        logger.error("Cannot generate CAPTCHA: %s" % e)
        raise Http500("Cannot generate CAPTCHA...")


